/*

  Classe SparseMatrix.
  Consente di definire una matrice sparsa e risolvere
  un sistema Ax = b utilizzando il metodo delle
  iterazioni di Jacobi.

 */

#include <iostream>     // Per cout in try/catch!
#include <cmath>        // Per funzione sqrt()!
#include <stdlib.h>     // Per funzione exit!
#include "Jacobi.h"
using namespace std;

// costruttore della classe

SparseMatrix::SparseMatrix( long int N, int M )
{
  dim = N;
  numDiags = M;
  // Vettore temporaneo per iterazioni di Jacobi
  temp = new double[N];
  // Termini diagonali della matrice
  diag = new double[N];
  // Vettore della lista elementi fuori diagonale della matrice != 0
  mat = new listElement[N];
  for ( long int i = 0; i < dim; i++ )
    {
      mat[i].num = 0;
      mat[i].cols = new long int[numDiags];
      mat[i].vals = new double[numDiags];
    }
}

// distruttore della classe

SparseMatrix::~SparseMatrix( )
{
   delete temp;
   delete diag;
   for ( long int i = 0; i < dim; i++ )
     {
       delete mat[i].cols;
       delete mat[i].vals;
     }
   delete mat;
}

// setta la diagonale della matrice, per elemento j-esimo

void SparseMatrix::setDiagonal( long int i, double val )
{
  if ( ( 0 <= i ) && ( i < dim ) ) diag[i] = val;
}


// setta elemento i,j fuori diagonale

void SparseMatrix::setOffDiagonal( long int i, long int j, double val )
{
  if ( ( 0 <= i ) && ( i < dim ) && ( 0 <= j ) && ( j < dim ) )
    {
      int n = mat[i].num;
      try {
	if ( n >= numDiags ) throw "Errore: la larghezza di banda selezionata non e' sufficiente...";
      }
      catch ( char const *str ) {
	cout << str << endl;
	cout << "Aumentare il secondo parametro nel costruttore della classe SparseMatrix..." << endl;
	exit( 0 );
      }
      mat[i].cols[n] = j;
      mat[i].vals[n] = val;
      mat[i].num++;
    }
}


// singola iterazione di Jacobi
void SparseMatrix::iterationStep( double *b, double *x, double &dist )
{
  double sumDist = 0.0;

  for ( long int i = 0; i < dim; i++ )
    {
      temp[i] = x[i];
      // somme dei termini fuori diagonale
      int n = mat[i].num;
      double sum = 0.0;
      for ( long int j = 0; j < n; j++ )
	{
	  long int jj = mat[i].cols[j];
	  sum += mat[i].vals[j] * temp[jj];
	}
      x[i] = ( b[i] - sum ) / diag[i];
      sumDist += ( temp[i] - x[i] ) * ( temp[i] - x[i] );
    }
  dist = sqrt( sumDist );
}
