/*

  Programma Rilassamento.cpp.
  risolve l'equazione di Poisson:
  d^2 V/dx^2 + d^2 V/dy^2 = g(x,y)
  con condizioni al bordo periodiche, in un 
  dominio rettangolare (x,y) = [a,b]x[c,d],
  con un metodo di rilassamento (di Jacobi).
  Utilizza la classe Jacobi.h, che costruisce
  una matrice sparsa e applica il metodo di
  Jacobi per invertire il sistema.
  La soluzione e' calcolata nel caso particolare:
  g(x,y) = -13.0*sin(2*x)*cos(3*y)
  che ammette per soluzione:
  V(x,y) = sin(2*x)*cos(3*y)

 */

#include <iostream>
#include <fstream>
#include <cmath>
#include "Jacobi.h"
using namespace std;


main()
{
  cout << "Soluzione di una equazione di Poisson nabla^2(V) = g(x,y)" << endl;

  // Parametri di input

  cout << "Inserisci il numero di punti griglia nella direzione x (Nx): ";
  int Nx;
  cin >> Nx;
  cout << "Inserisci il numero di punti griglia nella direzione y (Ny): ";
  int Ny;
  cin >> Ny;
  cout << "Intervallo in x ([a,b]): ";
  double a, b;
  cin >> a >> b;
  cout << "Intervallo in y ([c,d]): ";
  double c, d;
  cin >> c >> d;
  cout << "Tolleranza nel calcolo della soluzione: ";
  double eps;
  cin >> eps;
  cout << "Nome del file di uscita: ";
  char fileName[250];
  cin >> fileName;

  // Definizione dei vettori per i punti griglia

  double dx = ( b - a ) / ( Nx - 1 );
  double dy = ( d - c ) / ( Ny - 1 );
  double *x = new double[Nx];
  double *y = new double[Ny];
  for ( int i = 0; i < Nx; i++ )
    x[i] = a + i * dx;
  for ( int j = 0; j < Ny; j++ )
    y[j] = c + j * dy;

  // Costanti utili

  double dx2 = dx * dx;
  double dy2 = dy * dy;
  double alpha = dy2 / ( dx2 + dy2 );
  double beta  = dx2 / ( dx2 + dy2 );
  double gam   = dx2 * dy2 / ( dx2 + dy2 );

  // Secondo membro e guess iniziale della soluzione.
  // Nota come la periodicita' impone che la soluzione
  // sul bordo esterno in x e y sia nota, cosicche'
  // il sistema effettivo risulta di dimensione:
  // ( Nx - 1 ) * ( Ny - 1 )

  long int Ntot = ( Nx - 1 ) * ( Ny - 1 );
  double *g = new double[Ntot];
  double *V = new double[Ntot];
  for ( int j = 0; j < Ny - 1; j++ )
    for ( int i = 0; i < Nx - 1; i++ )
      {
	long int k = i + j * ( Nx - 1 );
	g[k] = -13.0 * sin( 2.0 * x[i] ) * cos( 3.0 * y[j] );
	V[k] = g[k];
	g[k] *= gam;
      }

  // Istanza dell'oggetto matrice

  SparseMatrix matrice = SparseMatrix( Ntot, 4 );

  // Definizione degli elementi diagonali della matrice

  for ( long int j = 0; j < Ntot; j++ )
    {
      matrice.setDiagonal( j, -2.0 );
    }

  // Definizione degli elementi fuori-diagonale
  // La matrice viene suddivisa in sotto-matrici
  // ( (Ny-1)*(Ny-1) sotto-matrici di dimensione
  //   (Nx-1)*(Nx-1) ciascuna )

  long int istart, jstart;

  for ( int jmat = 0; jmat < Ny - 1; jmat++ )       // riga di sotto-matrici
    for ( int imat = 0; imat < Ny - 1; imat++ )     // colonna di sotto-matrici
      {
	// elementi di partenza della sotto-matrice considerata
	// rispetto alla matrice principale. Ciascuna sotto-matrice
	// ha dimensione (Nx-1)*(Nx-1)!

	istart = imat * ( Nx - 1 );
	jstart = jmat * ( Nx - 1 );

	// Nota: i termini della parte triangolare inferiore della matrice
	// sono sempre considerati i trasposti di quella supariore
	// in quanto la matrice e' simmetrica!

	if ( imat == jmat )      // sotto-matrici lungo la diagonale!
	  {
	    for ( int ii = 0; ii < Nx - 2; ii++ )
	      {
		// codiagonale superiore ed inferiore (termini alpha),
		// elemento i=j+1,j
		long int i = istart + ii + 1;
		long int j = jstart + ii;
		matrice.setOffDiagonal( i, j, alpha );
		matrice.setOffDiagonal( j, i, alpha );  // trasposta!
	      }

	    // angoli delle sotto-matrici (termini alpha)
	    long int i = istart + Nx - 2;
	    long int j = jstart;
	    matrice.setOffDiagonal( i, j, alpha );
	    matrice.setOffDiagonal( j, i, alpha );  // trasposta!
	  }   // fine if ( imat == jmat )

	// Diagonale piu' interna di beta (sulla diagonale della
	// sotto-matrice imat=jmat+1,jmat e trasposta)

	if ( imat == jmat + 1 )
	  {
	    for ( int ii = 0; ii < Nx - 1; ii++ )
	      {
		long int i = istart + ii;
		long int j = jstart + ii;
		matrice.setOffDiagonal( i, j, beta );
		matrice.setOffDiagonal( j, i, beta );  // trasposta!
	      }  // fine for ( int ii = 0; ... )
	  }   // fine if ( imat == jmat + 1 )

	// Diagonale piu' esterna di beta (sulla diagonale della
	// sotto-matrice imat=Ny-2,jmat=0 e trasposta)

	if ( ( imat == Ny - 2 ) && ( jmat == 0 ) )
	  {
	    for ( int ii = 0; ii < Nx - 1; ii++ )
	      {
		long int i = istart + ii;
		long int j = jstart + ii;
		matrice.setOffDiagonal( i, j, beta );
		matrice.setOffDiagonal( j, i, beta );  // trasposta!
	      }  // fine for ( int ii = 0; ... )
	  }  // fine if ( ( imat == Ny-2 ) && ... )

      }  // fine for ( int imat = 0; ... ), for ( jmat = 0; ... )

  // Fine definizione matrice

  // Iterazioni di Jacobi

  long int numIter = 0;
  double distanza = 1.e+20;

  while ( distanza >= eps )
    {
      numIter++;
      cout << "Iterazione n. : " << numIter << endl;
      matrice.iterationStep( g, V, distanza );
      cout << "media quadratica fra due iterazioni successive: " << distanza << endl;
    }  // Fine while ( distanza >= eps )

  // Fine iterazioni di Jacobi

  cout << "Numero di punti in x: " << Nx << endl;
  cout << "Numero di punti in y: " << Ny << endl;
  cout << "Tolleranza richiesta eps: " << eps << endl;
  cout << "Precisione raggiunta: " << distanza << endl;
  cout << "Numero di iterazioni richiesto: " << numIter << endl;
  cout << "Soluzione numerica ed esatta su file: " << fileName << endl;


  // Stampa del risultato

  ofstream fout;
  fout.open( fileName );
  fout.precision( 15 );
  for ( int j = 0; j < Ny - 1; j++ )
    {
      for ( int i = 0; i < Nx - 1; i++ )
	{
	  double Vex = sin( 2.0 * x[i] ) * cos( 3.0 * y[j] );
	  long int k = i + j * ( Nx - 1 );
	  fout << x[i] << "   " << y[j] << "   " << V[k] << "   " << Vex << endl;
	}  // Fine for ( long int i = 0; ... )
      fout << "   " << endl;
    }  // Fine for ( long int j = 0; ... )
  fout.close();

}   // Fine main()
